function predicates=FindCandidatePredicates(challenges,rtables,responses,predicates_all,H,display_prefix)
% predicates=FindCandidatePredicates(challenges,rtables,responses,predicates_all,H,display_prefix)
% Find valid predicates that fit the given challenges and responses.
% The output is a 1xnr cell matrix, whose element is a cell matrix, which
% stores all the valid predicates after i=1~nr rounds of challenges.
%
% Developer: Shujun Li @ www.hooklee.com, 2009.

if nargin<6
    display_prefix='';
    if nargin<5
        H=26;
        if nargin<3
            disp('FindCandidatePredicates: at least three input arguments are needed!');
            return;
        end
    end
end
if nargin<4
    predicates_all=[];
end

nr=numel(challenges);
[m n l]=size(challenges{1});

% Note: Loading all the predicates from a data file is even slower 
% than creating them again online.
predicates=cell(1,nr);
if isempty(predicates_all)
    fprintf('%sCreating all possible predicates for testing, please wait...\n',display_prefix);
    t0=cputime;
    % Use 1-D cell array to ease removal of invalid predicates.
    predicates{1}=CreateAllPossiblePredicates(m,n,H);
    fprintf('%sFinish creating all possible predicates. Time consumed: %s.\n',...
        display_prefix,Seconds2TimeStr(cputime-t0));
else
    predicates{1}=predicates_all;
end
predicates_tested_number=0;
fprintf('%sStart testing validity of predicates.\n', display_prefix);
t0=cputime;
for r=1:nr
    fprintf('%sRound %d: start testing validity of predicates.\n',display_prefix,r);
    if r>1
        predicates{r}=predicates{r-1};
    end
    predicates_r=predicates{r};
    challenge=challenges{r};
    response=responses{r};
    rtable=rtables{r};
    s=numel(predicates{r});
    % Using logical array is the KEY to drastically speedup the process!!
    % If we change to use a growing array, the speed can be tens of times
    % slower! So try to avoid growing matrix as possible as we can.
    % Experiments showed the latter runs 44 times slower than the former!
    InvalidIndices=false(1,s);
    predicates_total_number=predicates_tested_number+sum(round(s*4.^(0:-1:r-nr)));
    tr=cputime;
    for i=1:s
        % Avoid invoking GetHiddenResponse() and ismember() to speedup the
        % calculation process.
        predicate_ri=predicates_r{i};
        response_hidden1=1;
        response_hidden2=1;
		% Avoid using a two-level loop to further speedup the process.
		for il=1:l
            % Note 1: strfind can run much faster than ismember and find!
            % Note 2: 2^(il-1) runs much faster than bitset(0,il)!
            if ~isempty(strfind(challenge{predicate_ri(1,1),predicate_ri(1,2),il},predicate_ri(1,3)))
                response_hidden1=response_hidden1+2^(il-1);
            end
            if ~isempty(strfind(challenge{predicate_ri(2,1),predicate_ri(2,2),il},predicate_ri(2,3)))
                response_hidden2=response_hidden2+2^(il-1);
            end
		end
        % Compare the responses and record the indices corresponding to invalid predicates.
        response2=rtable{response_hidden1,response_hidden2};
        if (response(1)~=response2(1) || response(2)~=response2(2))
            InvalidIndices(i)=1;
        end
        % Display testing progress.
%         if rem(i,80000)==0
%             t1=cputime-t0;
%             percentage=100*(predicates_tested_number+i)/predicates_total_number;
%             t2=t1*(100-percentage)/percentage;
%             InvalidPredicateNumber=numel(find(InvalidIndices(1:i)==1));
%             fprintf('%sRound %d: %d/%d (%.2g%%) tested, %d (%.2g%%) passed; time: %s -- %s\n',...
%                 display_prefix,r,i,s,100*i/s,i-InvalidPredicateNumber,...
%                 100*(1-InvalidPredicateNumber/i),Seconds2TimeStr(t1),Seconds2TimeStr(t2));
%         end
    end
    % Update the number of tested predicates.
    predicates_tested_number=predicates_tested_number+s;
    % Delete all invalid predicates.
    predicates{r}(InvalidIndices)=[];
    InvalidPredicateNumber=numel(find(InvalidIndices==1));
    fprintf('%sRound %d: all %d predicates tested, %d (%.2g%%) passed; time consumed for this round: %s.\n',...
        display_prefix,r,s,s-InvalidPredicateNumber,100*(1-InvalidPredicateNumber/s),Seconds2TimeStr(cputime-tr));
end
fprintf('%sFinish testing validity of predicates: %d predicate(s) passed; time consumed: %s.\n', ...
    display_prefix,size(predicates,2),Seconds2TimeStr(cputime-t0));
