% This is the main file you should run.
% You can determine how many runs you want to simulate the attack.
% For each run, a password and len authentication sessions will be randomly
% generated, and then a probablistic attack will be launched for all the 
% authentication sessions to see how many sessions are needed to break the
% cell index pair and alos probably some secret characters.
%
% Developer: Shujun Li @ www.hooklee.com, 2009.

clc;
clear all;

% Ask the user to set number of runs.
nRuns=input('Please input the number of runs of attacks you want to simulate (<=0 - infinite): ');
if nRuns<=0
    nRuns=intmax('int32');
end

% Use random permutation mode as the default. 
% You can change it to 'R' -- pure random mode, in which the predicate
% index is randomly generated with possible reuse.
I_type='P'; 

% nr and l are parameters that may be changed very often.
nr=5;
l=2;
% m, n, H and len are parameters that may be changed less frequently.
m=5;
n=5;
H=26;
len=10;
len2=2*len;
% p cannot be changed, because only p=2 is supported currently!
p=2;

if I_type(1)=='R'
    fprintf('Please input the number of authentication sessions for each run (1~2*%d)', len);
    na=input(': ');
    na=round(na); % Make it an integer.
    if (na<1 || na>len2)
        na=len;
    end
else
    na=len; % In this mode, the value of na can NOT be changed.
end

fprintf('\nFor each run of attack, the program will first simulate %d authentication sessions.\n',len);
fprintf('Then, it will calculate the set P_i for each authentication session.\n');
fprintf('Next, it will simulate %d attacks with 1,...,%d observed authentication sessions.\n',len,len);
fprintf('Finally, it will show all the candidate passwords for each simulated attack.\n');
fprintf('  Candidate characters corresponding to the same position are put in a pair of brackets.\n');
fprintf('  Asterisks ''***'' denote secret characters which remain unbroken.\n');

fprintf('\nPress any key to start the simulations.\n');
pause;

t00=cputime;
for iRuns=1:nRuns
    t0=cputime;

    fprintf('\n===========Run %d===========\n',iRuns);

    % Generate a random password.
    password=GeneratePassword(m,n,H,len,p);
    fprintf('\nRandomly generated password:\n');
    DisplayPassword(password);

    predicates=cell(1,na);
    challenges=cell(1,na);
    rtables=cell(1,na);
    responses=cell(1,na);

    fprintf('\nStart generating %d random authentication sessions...\n',na);

    if I_type(1)=='R'
        I=myrand(len,1,na);
    else
        I=randperm(len);
    end
    for i=1:na
        fprintf('=====Authentication Session %d=====\n',i);
        % First generate a random index I and a predicate.
        predicates{i}=GeneratePredicate(password,I(i));
        fprintf('I = %d: predicate = %s\n', I(i),Predicate2Str(predicates{i}));
        % Generate a random challenge, a response table and a response.
        [challenges{i},rtables{i},responses{i}]=GenerateSession(predicates{i},nr,m,n,H,l,p);
    end

    fprintf('\nStart collecting data for the statistical attack...\n');

    predicates_sets=cell(na,nr);

    fprintf('Pre-compute all possible predicates for the system parameters.\n');
    predicates_all=CreateAllPossiblePredicates(m,n,H);

    fprintf('Find candidate predicates for each authentication session.\n');
    for i=1:na
        fprintf('=====Authentication Session %d=====\n',i);
        predicates_sets_na=FindCandidatePredicates(...
            challenges{i},rtables{i},responses{i},predicates_all,H);
        for j=1:nr
            predicates_sets{i,j}=predicates_sets_na{j};
        end
    end
    clear predicates_all predicates_sets_na; % Do not need to store these variables.

    fprintf('\nAll pre-computation finished. Now try to break the password...\n');

    PasswordsBroken=cell(1,na);
    % na_Password1Broken is the minimal number of observed sessions for the 
    % first successful attack.
    na_Password1Broken=0; % 0 = no success even with all na sessions.
    for i=1:na
        fprintf('Trying to break the password with %d observed authentication sessions...\n',i);
        PasswordsBroken{i}=BreakPassword(predicates_sets(1:i,nr),I);
        if (na_Password1Broken==0 && numel(PasswordsBroken{i})==1)
			na_Password1Broken_s=numel(PasswordsBroken{i});
			for j=1:na_Password1Broken_s
				if all(all(PasswordsBroken{i}{j}{1}==password(1:2,1:2)))
					na_Password1Broken=i;
					break;
				end
			end
        end
    end

    fprintf('Run %d finished! Total time consumed: %s.\n',...
        iRuns,Seconds2TimeStr(cputime-t0));

    clear t0 i j na_Password1Broken_s; % clear useless variables

    filename=sprintf('Data_%d_%s_%s',iRuns,...
        Subpassword2Str(password(1,:)),Subpassword2Str(password(2,:)));
    fprintf('Save all variables in a file named ''%s''.\n', filename);
    save(filename); % save all variables in a data file

    % Save the main statistics of each run.
    fprintf('\nSave the main statistics of each run.\n');
    Runs.nRuns=iRuns; % number of runs
    Runs.I{iRuns}=I;
    Runs.password{iRuns}=password;
    Runs.PasswordsBroken{iRuns}=PasswordsBroken;
    Runs.na_Password1Broken(iRuns)=na_Password1Broken;
    % Save results for each run, because the whole program may be 
    % forcly interupt by users at any time (by pressing <Ctrl>+C).
    % Note that a copy of Runs of different ime will also be stored in the 
    % data file of each authentication session.
    % The following one is mainly used for easing the global statistics.
    save Runs Runs;
end

fprintf('\nMaking statistics of the broken passwords...\n');
Runs.AttackSuccessRate=zeros(1,na);
Runs.BrokenPasswordNumber=zeros(1,na);
for i=1:na
    for iRuns=1:nRuns
        if all(all(Runs.PasswordsBroken{iRuns}{i}{1}{1}==Runs.password{iRuns}(1:2,1:2)))
            Runs.AttackSuccessRate(i)=Runs.AttackSuccessRate(i)+1;
        end
        Runs.BrokenPasswordNumber(i)=Runs.BrokenPasswordNumber(i)+numel(Runs.PasswordsBroken{iRuns}{i});
    end
end
Runs.AttackSuccessRate=Runs.AttackSuccessRate/nRuns;
Runs.BrokenPasswordNumber=Runs.BrokenPasswordNumber/nRuns;

% If the user breaks the running of the program, manually execute the
% following command to rename the global data file.
save(['Runs_' datestr(clock,30) '.mat'],'Runs');
delete Runs.mat;

fprintf('All runs finished! Total time consumed: %s.\n',Seconds2TimeStr(cputime-t00));

figure;
plot(1:na,Runs.AttackSuccessRate,'-*');
xlabel('Number of observed sessions');
ylabel('Success rate of breaking the secret cell index pair');
axis tight;
axis([1 10 0 1]);
grid;
figure;
plot(1:na,Runs.BrokenPasswordNumber,'-*');
xlabel('Number of observed sessions');
ylabel('Average number of broken secret cell index pairs');
axis tight;
axis([1 10 1 ceil(max(Runs.BrokenPasswordNumber))])
grid;
